(* ::Package:: *)

BeginPackage["CTMCDist`"]


CTMC::usage="CTMC[tau, ell, rho, absQ] is the data structure for a continuous-time Markov chain, where \n
 - tau: S \[Backslash] A \[RightArrow] \[ScriptCapitalD](S) is the probability transition function, \n
 - ell: S \[RightArrow] L is the labeling function, and \n
 - rho: S \[Backslash] A \[RightArrow] \[DoubleStruckCapitalR] is the exit-rate function, \n
 - absQ: S \[RightArrow] {True, False} is the absorbing-state characteristic function.";


AllPairs::usage="AllPairs[<states>] returns the list of all pairs of states";


CTMCQ::usage="CTMCQ[<CTMC>] checks whether the given is a well-defined continuous-time Markov Chain with rational transition function and exit-rates function.";


CTMCtf::usage="CTMCtf[<array rules>, <states>] returns a transition function out of an array rules describing the transition matrix.";


PlotCTMC::usage="PlotCTMC[<states>, <CTMC>] displays the underlying graph of the continuous-time Markov chain <CTMC> on the states <states>.";


BDistCTMC::usage="BDistCTMC[<CTMC>, <discount factor>, <query list> [, Options]] computes the bisimilarity distance between all the pairs of states in <query list>.";


IterativeAlg::usage="IterativeAlg[<states>, <CTMC>, <discount factor> [, Options]] computes an under-approximation of the bisimilarity distance between all the pairs of states using the iterative method. \n
Available options are: \n
- IterationLimit (default value Infinity) if set to n, the number of iterations will be limited to n; \n
- Timeout (default value Infinity) if set to t, the computation will stop after execeeding t seconds. The value returned will correspond to that returned by the last performed iteration.";


Timeout::usage="Option for IterativeAlg (default value Infinity); if set to t, the computation will stop after execeeding t seconds. The value returned will correspond to that returned by the last performed iteration.";


IterationLimit::usage="Option for IterativeAlg (default value Infinity); if set to n the number of iterations will be limited to n.";


LPMethod::usage="LPMethod[<states>, <CTMC>, <discount factor> [, Options]] 
computes the bisimilarity distance between all the pairs of states as the solution of a Linear Program. \n
Available options are:\n
- Method (dafault \"InteriorPoint\") the method used to solve the Linear Program (see LinearProgramming function for more details); \n
- Timeout (default value Infinity) if set to t, the computation will be aborted after execeeding t seconds.";


Verbose::usage="Option for BDistCTMC (default value False). If set to True, during the computation some information about the current coupling structure is displayed.";


Output::usage="Option for BDistCTMC (default value RulesForm); if set to RulesForm the output distance will di displaied as a set of rules that associate with each pair in the input query its respective distance value; if set to VerboseForm, the output will display some information about the computation.";


ConsistencyCheck::usage="Option for BDistCTMC (default value True); if set to True, a sanity check of the input is performed before the actual algorithm starts.";


Estimates::usage="Option for BDistCTMC (default value {}). An over-estimates for the pair of states {s1,s2} can be given in the form '{s1,s2} -> e'. Then, the on-the-fly strategy will consider 'e' as the exact distance value for the pair '{s1,s2}'";


BisimQuotientCTMC::usage="BisimQuotientCTMC[<states>, <CTMC>] returns the pair {Q, M} where Q is the set of the quotient of S w.r.t. bisimilarity and M the CTMC that represents the bisimilarity quotient of the input CTMC.";


BisimClassesCTMC::usage="BisimClassesCTMC[<states>, <CTMC>] returns the partitioning of <states> w.r.t. bisimilarity equivalence.";


RandomCTMC::usage="RandomCTMC[<states>, <outer-degree>, <absorbing-degree>] randomly generates an CTMC with given out-degree and absorbing-degree.";


Begin["`Private`"]


(* ::Subsection:: *)
(*Continuous-Time Markov Chains (CTMCs)*)


AllPairs[S_]:=Tuples[S,2];


MCQ[MC[TM_,labels_]]:=
Module[{size,a,b},
size=Length@labels; {a,b}=Dimensions[TM];
(* check if the sizes are consistent, and if TM is a probability distribution matrix *)
size==a\[And]a==b\[And] TestAll[(Plus@@#&)/@TM,#==1&]
]/;MatchQ[labels,{__Integer}]\[And] (* labels is a list of integers *)
MatrixQ[TM,MatchQ[#,_Rational]\[Or] IntegerQ[#]&](* TM is a matrix of rationals *);
MCQ[_]:=False;


JoinMC[seq__MC]:=Fold[JoinMCAux,First@{seq},Rest@{seq}];
JoinMCAux[MC[tm1_,l1_],MC[tm2_,l2_]]:=
Module[{n = Length@l1, m=Length@l2},
MC[ArrayFlatten[{{tm1,SparseArray[{},{n,m}]},{SparseArray[{},{m,n}],tm2}}],
Join[l1,l2]]
];


(* ::Text:: *)
(*PlotMC returns a graphical representation of a given Markov chain *)


PlotCTMC[S_,CTMC[tau_,ell_,rho_,absQ_]]:=
Module[{f,g,edgelist,vertexlst,vrender,Unbox,NAbsS},
NAbsS = Select[S,Not@absQ@#&];

Unbox[FractionBox[a_,b_]]:=ToString[a / b];
Unbox[a_]:= ToString[a];

f[x_][y_->p_]:= Labeled[x->y,InputForm@p];
vrender[coord_,name_,{w_,h_}]:= {Hue[ell[name]],Disk[coord,0.1],Black, Scale[Text[Grid[{{InputForm@ell[name]}}],coord],0.1]}/; absQ[name];
vrender[coord_,name_,{w_,h_}]:= {Hue[ell[name]],Disk[coord,0.1],Black, Text[Grid[{{InputForm@ell[name]},{InputForm@rho[name]}},Dividers->Center,Alignment->{{Top},{Bottom}}],coord]}/; Not@absQ[name];

(*panelLabel[s_]:=Tooltip[s,TableForm[{{"Label:",Framed@ell[s]}}]]/; absQ[s];
panelLabel[s_]:=Tooltip[s,TableForm[{{"Label:",Framed@ell[s]},{"Rate:",Framed@rho[s]}}]]/; Not@absQ[s];

vertexlst = (# -> Placed["Name",Center,panelLabel])&/@S;*)

edgelist= Flatten[(f[#]/@tau[#])&/@NAbsS];

Graph[edgelist,VertexShapeFunction->vrender](**)
];


(* ::Subsection::Closed:: *)
(*Random generation of CTMCs*)


RandomRational[rmax_,n_]:=
(RandomInteger[{1,#}]/#)&/@RandomInteger[{1,rmax},n]


(* ::Text:: *)
(*RandomProbDistr gives a pseudorandom list of rationals lst = {r1, ..., rn} such that Total[lst] == 1. *)


RandomProbDistr[rmax_,n_]:=
Module[{rnd,total},
rnd=RandomRational[rmax,n]; total=Total[rnd];
(#/total)&/@rnd ];


RandomTransitionFunction[States_,maxDegree_]:=
Module[{\[Tau],degree,successors,probabilities,len=Length[States]},
Do[
degree = RandomInteger[{1,maxDegree}];
successors = RandomSample[States,degree];
probabilities = RandomProbDistr[len,degree];
\[Tau][s] = Thread[Rule[successors,probabilities]];
,{s,States}];
Return[\[Tau]] 
]/;Length[States]>=maxDegree;
RandomTransitionFunction[numStates_,maxDegree_]:=
(Message[RandomTransitionFunction::nnarg,numStates,maxDegree];Abort[]);


RandomTransitionFunction::nnarg="The argument `1` is not greater than or equal to `2`.";


RandomAbsorbing[States_,numAbs_]:=
Module[{AbsQ},
AbsQ[_]:=False;
Do[AbsQ[s]=True,{s,RandomSample[States,numAbs]}];
Return[AbsQ]
];


RandomLabeling[States_]:=
Module[{ell,L,R,len=Length[States]},
L=RandomRational[len,RandomInteger[{1,len}]];
R=RandomChoice[L,len];
Do[ell[States[[i]]]= R[[i]],{i,len}];
Return[ell]
];


RandomRates[States_]:=
Module[{rho,L,R,len=Length[States]},
L=RandomInteger[{1,20},RandomInteger[{1,len}]];
R=RandomChoice[L,len];
Do[rho[States[[i]]]= R[[i]],{i,len}];
Return[rho]
];


RandomCTMC[States_,maxDegree_,numAbs_]:=
Module[{S,tau,ell,rho,absQ},
absQ = RandomAbsorbing[States,numAbs];
S = Select[States,Not@absQ@#&];
tau = RandomTransitionFunction[S,maxDegree];
ell = RandomLabeling[States];
rho = RandomRates[S];
CTMC[tau,ell,rho,absQ]
];


(* ::Subsection::Closed:: *)
(*LP Utilities*)


GetConstrMatrix[inequalities_, vars_]:=
Module[{ConstrRow,rows,consts},
ConstrRow[inequality_]:=
Module[{b,A},
{b,A}=CoefficientArrays[Equal@@inequality,vars];
{A,{-b,Head[inequality]/.{LessEqual->-1,Equal->0,GreaterEqual->1}}}
];
ArrayFlatten/@Thread[ConstrRow/@inequalities]
];


(* ::Subsection::Closed:: *)
(*Utilities*)


labelDist[l1_, l2_]:=Quiet@N@Norm[l1-l2,Infinity];


rateDist[\[Rho]1_Integer,\[Rho]2_Integer]:=0 /;\[Rho]1==\[Rho]2;
rateDist[\[Alpha]_Integer,\[Lambda]_Integer]:=Quiet@N@Abs[ (\[Alpha]/\[Lambda])^(-(\[Alpha]/(\[Alpha]-\[Lambda])))-(\[Alpha]/\[Lambda])^(-(\[Lambda]/(\[Alpha]-\[Lambda])))];


DebugPrint[s___]:=Print[s]/;OptionValue[BDistCTMC,Verbose];


PrPrint[Schedule[mapS_,mapT_,basicCells_,V_]]:=
Module[{basic},
basic[value_,index_]:=Style[EpsilonRemoval@value,Red,Bold]/;MemberQ[basicCells,index];
basic[value_,_]:=value;
TableForm[MapIndexed[basic,V,{2}],TableHeadings->{ToString[#[[2]]]&/@mapS,ToString[#[[2]]]&/@mapT}]
];
PrPrint[x_]:=x;


PrintActiveProblems[dist_,coupling_,queryLst_]:=
(Print["Reachable states from ",queryLst, " in the current coupling"];
Print["Currently demanded pairs: ",ActivePairs[dist,coupling,queryLst]];
Scan[Print[# -> PrPrint[coupling[#]]]&,ActivePairs[dist,coupling,queryLst]])/;OptionValue[BDistCTMC,Verbose];


(* ::Subsection::Closed:: *)
(*Transportation Problem*)


scheduleCost[cost_,schedule:Schedule[_,_,basicCells_,V_]]:=
Total[Extract[cost,#]*EpsilonRemoval[Extract[V,#]]&/@ basicCells];


cell2pair[{s_,t_},Schedule[mapS_,mapT_,_,_]]:={s/.mapS,t/.mapT};


(* ::Text:: *)
(*Given two \[Epsilon]-numbers x and y, EpsilonLeq[x,y] determines if x <= y provided that \[Epsilon] -> +0.*)


EpsilonLeq[x_,y_]:=
If[EpsilonRemoval[x-y]==0,(* check if x-y depends only on \[Epsilon]*)
EpsilonFix[x-y,1]<= 0,(* set \[Epsilon] > 0 and check again *)
EpsilonRemoval[x-y]<= 0 (* set e[]= 0 and check again *)
];


EpsilonMin[x_,y_]:= If[EpsilonLeq[x,y],x,y];


EpsilonPerturbation[\[Pi]s_,\[Pi]t_]:=
Module[{AddEpsilon},
AddEpsilon[k_][{a_,id_}]:={a + k e[],id};
{AddEpsilon[1]/@\[Pi]s, Append[Most[\[Pi]t],AddEpsilon[Length@\[Pi]s][Last@\[Pi]t]]}
];


EpsilonFix[x_,\[Epsilon]value_]:=x/.e[]->\[Epsilon]value;


EpsilonRemoval[x_]:=EpsilonFix[x,0];


indexed[lst_]:=Thread[{lst,Table[i,{i,1,Length@lst}]}];


GuessVertexAux[_][\[Pi]s_,\[Pi]t_]:={}/;\[Pi]s=={}||\[Pi]t=={};
GuessVertexAux[strategy_][\[Pi]s_,\[Pi]t_]:=
Module[{G},
G[{i_,j_}]:=
Module[{rec},
rec=If[EpsilonLeq[\[Pi]s[[i,1]],\[Pi]t[[j,1]]],(*\[Pi]s\[LeftDoubleBracket]i,1\[RightDoubleBracket] \[LessEqual] \[Pi]t\[LeftDoubleBracket]j,1\[RightDoubleBracket]*)
GuessVertexAux[strategy][Drop[\[Pi]s,{i}],ReplacePart[\[Pi]t,{j,1}->\[Pi]t[[j,1]]- \[Pi]s[[i,1]]]],
GuessVertexAux[strategy][ReplacePart[\[Pi]s,{i,1}->\[Pi]s[[i,1]]- \[Pi]t[[j,1]]],Drop[\[Pi]t,{j}]]];
Prepend[rec,{\[Pi]s[[i,2]],\[Pi]t[[j,2]]}->EpsilonMin[\[Pi]s[[i,1]],\[Pi]t[[j,1]]]]
];
G[strategy[\[Pi]s,\[Pi]t]]
];


NorthWestStrategy[_,_]:={1,1};


ToMatrix[\[Pi]s_,\[Pi]t_,vertex_]:=
Module[{mapS,revmapS,mapT,revmapT,toMap,mapIndex,mappedvertex},
toMap[indexlst_]:=
Module[{index=Table[i,{i,1,Length@indexlst}] },
{Thread[Rule[index,indexlst]],Thread[Rule[indexlst,index]]}];
mapIndex[{s_,t_}->val_]:={s/.revmapS,t/.revmapT}->val;

{{mapS,revmapS},{mapT,revmapT}} = toMap[#[[All,2]]]&/@{\[Pi]s,\[Pi]t};
mappedvertex= mapIndex/@vertex;
Schedule[mapS,mapT,mappedvertex[[All,1]],SparseArray[mappedvertex,Length/@{\[Pi]s,\[Pi]t}]]
];


GuessVertex[\[Pi]s_,\[Pi]t_]:=
Module[{\[Epsilon]\[Pi]s,\[Epsilon]\[Pi]t},
{\[Epsilon]\[Pi]s,\[Epsilon]\[Pi]t}=EpsilonPerturbation[\[Pi]s, \[Pi]t];
ToMatrix[\[Epsilon]\[Pi]s,\[Epsilon]\[Pi]t,GuessVertexAux[NorthWestStrategy][\[Epsilon]\[Pi]s,\[Epsilon]\[Pi]t]]
];


(* ::Text:: *)
(*AddtoBase[d, vertex] == {reducedcost, non-basic cell} *)
(*returns the non-basic cell which has the minimal reduced cost, if the vertex has no non-basic cells, then it returns {\[Infinity], {0,0}}.*)


AddtoBase[costMatrix_,Schedule[mapS_,mapT_,basicCells_,V_]]:=
Module[{sizeS=Length@mapS, sizeT=Length@mapT, 
t,getCost,ReducedCost,uv,makeRow,base},

makeRow[{1,j_}]:={SparseArray[{{1,sizeS+j}->1},{1,sizeS+sizeT}]};
makeRow[{i_,j_}]:={SparseArray[{{1,i}->1,{1,sizeS+j}->1},{1,sizeS+sizeT}]};

getCost[{i_,j_}]:=costMatrix[[i,j]];

ReducedCost[_,{i_,j_}]:=
Module[{reducedcost=Chop[getCost[{i,j}]-uv[[i]]-uv[[sizeS+j]],10^-5]},
If[base[[1]]>reducedcost,
base={reducedcost,{i,j}} (* update the current least reduced cost *)
];reducedcost];

(* u-v modifiers (Subscript[u, 1],...,Subscript[u, sizeS],Subscript[v, 1],...,Subscript[v, sizeT]) *)
uv=LinearSolve[Normal@ArrayFlatten[makeRow/@basicCells],getCost/@basicCells];
(*Print["u-v Problem:",MatrixForm@ArrayFlatten[makeRow/@basicCells],
MatrixForm@uv,MatrixForm[getCost/@basicCells]];*)
base={\[Infinity],{0,0}};(* init cell *)
(* scan the cells computing their reduced cost *)
MapIndexed[ReducedCost,V,{2}];
base (* return the cell with the least reduced cost *)
];


(* ::Text:: *)
(*SteppingStone[non-basic cell, vertex] returns the vertex obtained by inserting the given non-basic cell into the base *)


SteppingStone[{s_,t_},vertex:Schedule[mapS_,mapT_,basicCells_,V_]]:=
Module[{M,d,oldBasicCell,newBasicCells},
{oldBasicCell,M}=SteppingStonePath[{s,t},vertex];
(*DebugPrint[MatrixForm@M];*)
(*DebugPrint["old basic cell: ",oldBasicCell, ", new basic cell: ",{s,t}];*)
d= Extract[V,oldBasicCell];
(*DebugPrint["ammount moved: ",d];*)
newBasicCells = basicCells/.oldBasicCell->{s,t};
(*DebugPrint["Old basic cells:",basicCells,", new basic cells",newBasicCells];*)
Schedule[mapS,mapT,newBasicCells,SparseArray[V+d M]]
];


SteppingStonePath[{s_,t_},vertex:Schedule[mapS_,mapT_,basicCells_,V_]]:=
Module[{neighbors,visit,parent,cell,path,getMatrix,visited,f},

visited[_]:=False;
neighbors[{u_,True}]:={#[[2]],False}&/@(Select[basicCells,#[[1]]==u \[And] Not[visited[{#[[2]],False}]]&]);
neighbors[{v_,False}]:={#[[1]],True}&/@(Select[basicCells,#[[2]]==v \[And] Not[visited[{#[[1]],True}]]&]);

f[{{a_,True},{b_,False}}]:={a,b};
f[{{a_,False},{b_,True}}]:={b,a};

visit[node_]:=
(visited[node]=True;
Scan[(parent[#]=node; visit@#)&,neighbors[node]]
);

getMatrix[{},sign_]:={};
getMatrix[lst_,sign_]:=
(If[sign<0\[And] EpsilonLeq[Extract[V,First@lst] ,Extract[V,cell]],cell=First@lst];(* update cell *)
Append[getMatrix[Rest@lst,- sign], First@lst -> sign]); (* compute arrayrules *)

(*DebugPrint[PrPrint@vertex];*)

(* search for an alternating path from s to t *)
parent[{s,True}]={};(* {s,True} is the root of the tree *)
visit[{s,True}]; (* compute the spanning tree rooted in {s,True} *)

(* get the alternating path from {t,False} to {s,True} *)
path=Append[f/@Partition[(Most@NestWhileList[parent,{t,False},#!={}&]),2,1],{s,t}];
cell=First@path; path=getMatrix[path,-1];

{cell,SparseArray[path,Dimensions@V]}
];


BestVertex[costMatrix_,{redCost_,nbCell_},vertex_]:=
Module[{currVertex=vertex, currRedCost=redCost,currNbCell= nbCell},
While[currRedCost<0,
currVertex= SteppingStone[currNbCell,currVertex];(* move to a better vertex *)
(*DebugPrint[PrPrint@currVertex];*)
(* update the current least reduced cost *)
{currRedCost,currNbCell}= AddtoBase[costMatrix,currVertex];
];
currVertex
];


(* ::Subsection::Closed:: *)
(*Iterative Method (approximated)*)


(* ::Text:: *)
(*Given the numer of states |S|, BottomDist[|S|] returns the bottom distance*)


BottomDist := Module[{d},d[_,_]:=0; d]; 
TopDist := Module[{d},d[_,_]:=1; d]; 


(* ::Text:: *)
(*The funtion \[CapitalDelta] implements the distance transformer*)


\[CapitalDelta][CTMC[_,_,_,absQ_],\[Lambda]_][dist_][s_,t_]:= 1 /; absQ[s] != absQ[t];
\[CapitalDelta][CTMC[_,ell_,_,absQ_],\[Lambda]_][dist_][s_,t_]:= labelDist[ell[s],ell[t]] /; absQ[s] && absQ[t]; 
\[CapitalDelta][CTMC[tau_,ell_,rho_,absQ_],\[Lambda]_][dist_][s_,t_]:=
Module[{K,T},
(* Kantorovich distance *)
K[d_][u_,v_]:= Module[{schedule,Marginal,cost},
Marginal[state_]:= Cases[tau[state],HoldPattern[index_->val_/;val>0]->{val,index}];
schedule=GuessVertex[Marginal[u],Marginal[v]];
cost = RelativeCostMatrix[d,schedule];
schedule=BestVertex[cost,AddtoBase[cost,schedule],schedule];
scheduleCost[cost,schedule] ];
(* step distance *)
T[d_][u_,v_]:= rateDist[rho[u],rho[v]] + (1 - rateDist[rho[u],rho[v]]) * K[d][u,v];

Max[labelDist[ell[s],ell[t]], \[Lambda] * T[dist][s,t]]
]/; Not[absQ[s]] && Not[absQ[t]];


IterativeAlg[S_,ctmc:CTMC[tau_,ell_,rho_,absQ_],\[Lambda]_,
	OptionsPattern[{Timeout -> Infinity,Verbose->False,IterationLimit -> Infinity,Output->RulesForm}]]:=
Module[{t=0,iterCount=0,u,v,dt,old,curr,neq,allPairs=Tuples[S,2]},
(* initialization *)
old = TopDist; curr = BottomDist;
(* iteratively apply \[CapitalDelta], until the extact result or 
the Timeout has been reached *)
neq=True;
While[t <= OptionValue[Timeout] \[And] iterCount <= OptionValue[IterationLimit] \[And] neq,
Do[{u,v}=pair; old[u,v] = curr[u,v],{pair,allPairs}];
neq = False;
{dt,curr}= Timing[
Do[{u,v}=pair; 
curr[u,v] = \[CapitalDelta][ctmc,\[Lambda]][old][u,v];
neq = neq || curr[u,v] != old[u,v];
,{pair,allPairs}];
curr
];
t+=dt;iterCount++;
If[OptionValue[Verbose],
Print["Elapsed time:",t,", iteration:"iterCount];
Print["Current values", (#->getDistValue[curr,#]&)/@allPairs];
]
];
Switch[OptionValue[Output],
RulesForm, (#->getDistValue[curr,#]&)/@allPairs,
VerboseForm, R[t,iterCount,curr],
_,curr]
];


(* ::Subsection::Closed:: *)
(*Linear Program Characterization (exact)*)


LPMethod[S_,ctmc:CTMC[tau_,ell_,rho_,absQ_],\[Lambda]_,
	OptionsPattern[{Method -> "InteriorPoint",Timeout -> Infinity,Verbose->False,Output->RulesForm}]]:=
Module[{allPairs=Tuples[S,2],NA , NotAbsS,\[Tau],
alphavars,distvars,mvars,vars,(* variables *)
fluxc,distc,T,allconstr,objDist,(* explicit constraints *)
c,A,b,bounds (* LP coefficients *),
sol,t,dist,K,dL},

NotAbsS[{s_,t_}]:=Not[absQ[s]]\[And]Not[absQ[t]];
\[Tau][s_,t_]:=t/.Append[tau[s],_->0];
dL[{s_,t_}]:=labelDist[ell[s],ell[t]];

NA=Select[allPairs,NotAbsS];

alphavars= Flatten@Table[If[NotAbsS[{s,t}],y[u][s,t],{}],{s,S},{t,S},{u,S}]; 
distvars=d@@#&/@allPairs;
mvars=m@@#&/@NA;
vars=Join[distvars,alphavars,mvars];

fluxc[{s_,t_},{u_,v_}]:= y[u][s,t] - y[v][s,t] <= d[u,v];

(*distc[{s_,s_}]:=d[s,s]\[Equal]0;*)
distc[{s_,t_}]:=d[s,t]==1 /;Not[absQ[s]]==absQ[t];
distc[{s_,t_}]:=d[s,t]== labelDist[ell[s],ell[t]]/; absQ[s]\[And]absQ[t];
distc[{s_,t_}]:={
d[s,t]== labelDist[ell[s],ell[t]]+ T[{s,t}]-m[s,t],
m[s,t]<= labelDist[ell[s],ell[t]], m[s,t]<= T[{s,t}]
} /;NotAbsS[{s,t}];

T[{s_,t_}]:=Module[{rdist=rateDist[rho[s],rho[t]]},\[Lambda](rdist + (1-rdist)*K[{s,t}])];
K[{s_,t_}]:=Total[Table[(\[Tau][s,u]-\[Tau][t,u])*y[u][s,t] ,{u,S}]] ;

TimeConstrained[
{t,sol}= 
Timing[
allconstr = Flatten[{
Total[d@@#&/@NA] == Total[(T[#]+dL[#]-m@@#)&/@NA],
distc/@allPairs,
Table[fluxc[p1,p2],{p1,NA},{p2,allPairs}]
}];

objDist=Total[(K[#]+m@@#)&/@NA];

c =CoefficientArrays[objDist,vars][[2]];
{A,b}=GetConstrMatrix[allconstr,vars];
(* bounds *)
bounds=Join[Table[{0,1},{Length@distvars}],Table[{-Infinity,Infinity},{Length@alphavars}]]; 
Quiet@Chop[LinearProgramming[-c,A,b,Method->OptionValue[Method]],10^-5]
(*,Method\[Rule]"RevisedSimplex",Method\[Rule] "Simplex",Method\[Rule]"InteriorPoint"*)
];

(*Print[TableForm[distc/@allPairs]];
Print[TableForm[Thread[Rule[vars,sol]]]];*)
Module[{u,v},
Do[{u,v}=allPairs[[i]];dist[u,v] = sol[[i]];,{i,Length@allPairs}]
];

Switch[OptionValue[Output],
RulesForm,(#->dist@@#&)/@allPairs ,
VerboseForm,R[t,Null,dist],
_,dist],
OptionValue[Timeout],Print["The timeout of ",OptionValue[Timeout]," seconds has been exceeded."]]
];


(* ::Subsection::Closed:: *)
(*On-the-fly Algorithm for computing Bisimilarity Pseudometric (exact)*)


FixedDistanceQ[dist_,pair:{_,_}]:= MatchQ[dist@@(Sort@pair),_Fixed] ;(* says if the distance is fixed or not *)


VisitedPairQ[dist_,coupling_,pair:{_,_}]:= (* decide whether the distance pair has already been considered *)
FixedDistanceQ[dist,pair] \[Or] Not@MatchQ[coupling[pair],Undefined];


NextDemandedPairs[schedule:Schedule[mapS_,mapT_,basicCells_,V_]]:= 
DeleteDuplicates[Sort[cell2pair[#,schedule]]& /@ Select[basicCells,EpsilonRemoval[Extract[V,#]]>0&]];


AddPair[CTMC[_,_,_,_],dist_][coupling_,pair:{_,_}]:= coupling/; VisitedPairQ[dist,coupling,pair];
AddPair[ctmc:CTMC[\[Tau]_,_,_,_],dist_][coupling_,pair:{s_,t_}]:=
Module[{Marginal,newschedule},
Marginal[state_]:= Cases[\[Tau][state],HoldPattern[index_->v_/;v>0]->{v,index}];
(* guess a fesible TP schedule for each action *)
newschedule := GuessVertex[Marginal[s],Marginal[t]];
(* recursively update the coupling *)
UpdateCoupling[ctmc,dist][coupling,pair,newschedule]
];


AddPairs[ctmc:CTMC[_,_,_,_],dist_][coupling_,queryLst_]:=Fold[AddPair[ctmc,dist],coupling,queryLst];


UpdateCoupling[ctmc:CTMC[_,_,_,_],dist_][coupling_,pair:{_,_},schedule_Schedule]:=
(coupling[pair]=schedule; 
AddPairs[ctmc,dist][coupling,NextDemandedPairs[schedule]]);


ReachableProblems[dist_,coupling_,queryLst_]:=
Module[{visitedQ}, visitedQ[_]:= False; Fold[DFS[dist,coupling],visitedQ,queryLst]];
DFS[dist_,coupling_][visitedQ_,pair:{_,_}]:=visitedQ /; visitedQ@pair \[Or] FixedDistanceQ[dist,pair];
DFS[dist_,coupling_][visitedQ_,pair:{_,_}]:=
(visitedQ[pair]=True; Fold[DFS[dist,coupling],visitedQ,NextDemandedPairs[coupling[pair]]]);


ActivePairs[dist_,coupling_,queryLst_]:=
#[[1,1,1]]&/@Select[DownValues[Evaluate@ReachableProblems[dist,coupling,queryLst]],#[[2]]&];


UpdateDistanceValues[\[Lambda]_,ctmc:CTMC[_,ell_,\[Rho]_,_],dist_,coupling_,queryLst_]:=
Module[{f,\[CapitalGamma],activeLst,fixedpairs = {},
xVars,xVals,updateVal},

updateVal[d_,pair:{s_,t_} -> val_]:=(d[s,t]=Fixed[0];AppendTo[fixedpairs,pair];Unset[coupling[pair]];d)/;val==0;
updateVal[d_,pair:{s_,t_} -> val_]:=(d[s,t]=Fixed[val];AppendTo[fixedpairs,pair];Unset[coupling[pair]];d)
	/;Chop[Abs[labelDist[ell[s],ell[t]]-val],10^-9]==0;
updateVal[d_,pair:{s_,t_} -> val_]:=(d[s,t]=val;d);

f[pair:{s_,t_}]:= (* constraints for d[s,t] *)
Module[{\[Alpha] = rateDist[\[Rho][s],\[Rho][t]]},
(* linear constraints to handle max{Subscript[d, L](\[ScriptL](s),\[ScriptL](t)), \[Lambda] \[ScriptCapitalT](d)(s,t)}*)
{x@@pair >= labelDist[ell[s],ell[t]], (* handler for Subscript[d, L](\[ScriptL](s),\[ScriptL](t)*)
x@@pair >= \[Lambda] (\[Alpha] + (1 - \[Alpha]) \[CapitalGamma][coupling[pair]])}]; (* handler for \[Lambda] \[ScriptCapitalT](d)(s,t) *)

\[CapitalGamma][schedule:Schedule[_,_,basicCells_,V_]]:=
Module[{var},
var[cell:{_,_}]:= Module[{pair=cell2pair[cell,schedule]},
	If[FixedDistanceQ[dist,pair],getDistValue[dist,pair],x@@(Sort@pair)]];
Total[var[#]*EpsilonRemoval[Extract[V,#]]&/@basicCells]
];

(* get the list of active pairs *)
activeLst=ActivePairs[dist,coupling,queryLst];
(* set the LP that characterizes the discrepancy assiciated with the actual coupling *)
xVars = x@@#&/@activeLst;(* value variables *)
xVals = Module[{A,b,c,sol},
{A,b}=GetConstrMatrix[Flatten[f/@activeLst],xVars]; (*contraint matrix and constant vector *)
c = CoefficientArrays[Total@xVars,xVars][[2]]; (* costs vector *)
(* solve the value problem for the current coupling  *)
sol= LinearProgramming[c,A,b,Method->"InteriorPoint"];(*"RevisedSimplex"*)
Chop[sol,10^-5]
];
(* update distances with new pair values *)
dist=Fold[updateVal,dist,Thread[Rule[activeLst,xVals]]];
DebugPrint["Current update:",Thread[Rule[activeLst,xVals]]];
DebugPrint["Pair fixed during the cuccent update: ",fixedpairs];
Return[dist]
];


EmptyCoupling[]:= Module[{C},C[_]:=Undefined; Return[C]];


getDistValue[dist_,pair:{_,_}]:=
Module[{getValue},
getValue[Fixed[v_]]:=v;
getValue[v_]:=v;

getValue[dist@@(Sort@pair)]
];


InitDistances[ctmc:CTMC[_,ell_,_,absQ_],estimates_]:=
Module[{dist,diffIndex,fixEstimate,TOP=1},

dist[s_,s_]:= Fixed[0]; (* {s,s} pairs are fixed to 0 *)
dist[s_,t_]:= Fixed[TOP] /; absQ[s] == Not[absQ[t]]; (* pairs are fixed to 1 *)
dist[s_,t_]:= Fixed[labelDist[ell[s],ell[t]]] /; absQ[s] \[And] absQ[t];
dist[_,_]:= TOP;

fixEstimate[pair:{_,_}->v_]:= (dist@@(Sort@pair)=v) /; v >= 0 && v <= 1 && Not@FixedDistanceQ[dist,pair];
Scan[fixEstimate,estimates]; (* set all the given estimates as fixed *)
Return[dist]
];


RelativeCostMatrix[dist_,schedule:Schedule[mapS_,mapT_,basicCells_,V_]]:=
Module[{getCellCost,cost,max},
getCellCost[_,cell_] := getDistValue[dist,cell2pair[cell,schedule]];
MapIndexed[getCellCost,V,{2}]
];


(* ::Text:: *)
(*MatchingToChange[coupling,dist,{s,t}] == {{u,v}, {redCost, nbcell}} if the schedule at (u,v) can be further minimized wrt TP with costs given by dist, otherwise MatchingToChange[coupling,dist,{s,t}] == {}.*)


MatchingToChange[coupling_,dist_,queryLst_]:=
Module[{Q,p,a,reducedcost,schedule,cell,siblings,visited},

visited[pair:{_,_}]:= FixedDistanceQ[dist,pair];

(* a new problem should not be visited if it has been visited already or 
its corresponding value is fixed *)
Q=Select[queryLst,Not@FixedDistanceQ[dist,#]&];
While[Length@Q >0,
(* take a problem from the queue and set it as already visited *)
p = First@Q; visited[p] = True;
(*DebugPrint["Visited problem:",{i,j}];*)
(* compute the least reduced cost for the current problem *)
schedule = coupling[p];
{reducedcost,cell} = AddtoBase[RelativeCostMatrix[dist,schedule],schedule];
If[reducedcost<0,
(* the first encoutered vertex to be changed *)
(*DebugPrint["Non optimal schedule for: ",p,
" reduced cost: ",reducedcost, " for introducing ",cell," as basic"];*)
Return[{p,{reducedcost,cell}}],
(* otherwise insert its siblings to the queue *)
siblings = Select[NextDemandedPairs[coupling[p]],Composition[Not,visited]];
Q = DeleteDuplicates@Join[Rest@Q,siblings]
]];
(* none of the reached problems can be further decreased *)
{}
];


Options[BDistCTMC]={Verbose->False,ConsistencyCheck->True,Output->RulesForm,Estimates-> {}};
BDistCTMC::query="The following queries are ill-formed: `1`.";
BDistCTMC::MarkovChain = "The 1st argument is not a well-formed rational-valued MDP.";
BDistCTMC::DiscountFactor ="The second argument, namely `1`, must be a rational in the interval (0,1)";
BDistCTMC::MultipleEstimates="There is a conflit between estimates (notice that if you have both 
{s,t}->v1 and {t,s}->v2, v1 must be equal to v2)";
BDistCTMC::WrongEstimates="Estimates must be a rationals or integers in [0,\[Infinity])";

(*BDistCTMC[ctmc:CTMC[_,_,_,_],\[Lambda]_, All,
options: OptionsPattern[{RulesForm->True,Verbose -> False,ConsistencyCheck->True,Estimates-> {}}]]:=
BDistCTMC[ctmc,\[Lambda],AllPairs[ctmc],options];*)
BDistCTMC[ctmc:CTMC[_,_,_,_],\[Lambda]_, queryLst_, 
options:OptionsPattern[{Output->RulesForm,Verbose -> False,ConsistencyCheck->True,Estimates-> {}}]]:=
(If[OptionValue[ConsistencyCheck],CheckInput[ctmc,\[Lambda], queryLst]];
(* Set Options for ComputeDistance *)
SetOptions[BDistCTMC,Output->OptionValue[Output]];
SetOptions[BDistCTMC,Verbose->OptionValue[Verbose]];
SetOptions[BDistCTMC,ConsistencyCheck->OptionValue[ConsistencyCheck]];
SetOptions[BDistCTMC,Estimates->OptionValue[Estimates]];
(* Start computing the Subscript[d, \[Lambda]](s,t) *)
ComputeDistancesAux[ctmc,\[Lambda],queryLst]
);


CheckInput[mdp:MDP[_,_,_],\[Lambda]_, queryLst_]:=
Module[{BadQuery,BadEstimate,sizeS=mdpDimensions[mdp][[1]],
badqueries,estimates},

BadQuery[size_][pair:{_Integer,_Integer}]:=Not[And@@(IntervalMemberQ[Interval[{1,size}],#]&/@pair)];
BadQuery[_][_]:=True;
BadEstimate[size_][pair_->value_]:=BadQuery[size][pair] \[Or] value<0\[Or]Not@MatchQ[value,_Rational|_Integer];

estimates = OptionValue[BDistCTMC,Estimates];
Which[
Not@MarkovChainQ[mdp],
	Message[BDistCTMC::MarkovChain];Abort[],
\[Lambda]<=0\[Or]\[Lambda]>=1\[Or]Not@MatchQ[\[Lambda],_Rational],
	 Message[BDistCTMC::DiscountFactor,\[Lambda]];Abort[],
Select[estimates,BadEstimate[mdpDimensions[mdp][[1]]]]>0,
	Message[BDistCTMC::WrongEstimates];Abort[],
Or@@(Length[#]!=1&/@(DeleteDuplicates[#[[All,2]]]&/@GatherBy[estimates,Sort@First@#&])),
	Message[BDistCTMC::MultipleEstimates];Abort[],
(badqueries =Select[queryLst,BadQuery[sizeS]])>0 ,
	Message[BDistCTMC::pair,badqueries];Abort[]
];
];


ComputeDistancesAux[ctmc:CTMC[\[Tau]_,ell_,\[Rho]_,absQ_], \[Lambda]_, queryLst_]:=
Module[{sortedQuery,query,FixPair,sizeA,sizeS, time,
currCoupling,currPolicy,currDist,empty,active,
vtc,scheduleIndex, rcCell,oldSchedule, newSchedule,iterations=0, return},

FixPair[dist_,{s_,t_}]:=
	(dist[s,t] = Fixed[dist[s,t]];
	Unset[currCoupling[{s,t}]]; 
	dist);

(* Initialization *)
time = First@Timing[
currDist=InitDistances[ctmc,OptionValue[BDistCTMC,Estimates]];
sortedQuery=DeleteDuplicates[Sort/@Select[queryLst,Not@FixedDistanceQ[currDist,#]&]];
While[sortedQuery != {},
DebugPrint["Pairs left to consider: ",sortedQuery];
(* pick at random a pair from the query *)
query=sortedQuery[[{1}]];(*sortedQuery\[LeftDoubleBracket]RandomInteger[{1,Length@sortedQuery},1]\[RightDoubleBracket];*)
DebugPrint["Current pair: ",query];
currCoupling = AddPairs[ctmc,currDist][EmptyCoupling[],query];
PrintActiveProblems[currDist,currCoupling,query];
currDist = UpdateDistanceValues[\[Lambda],ctmc,currDist,currCoupling,query];
(* look for a non-optimal transportation schedule *)
vtc = MatchingToChange[currCoupling,currDist,query];
While[vtc != {},
iterations++;
{scheduleIndex, rcCell} = vtc;(* consider the schedule to be optimized *)
DebugPrint["Coupling to change:",scheduleIndex];
(* move to a better coupling *)
oldSchedule= currCoupling[scheduleIndex];
newSchedule = BestVertex[RelativeCostMatrix[currDist,oldSchedule], rcCell, oldSchedule];
currCoupling = UpdateCoupling[ctmc,currDist][currCoupling,scheduleIndex, newSchedule];

PrintActiveProblems[currDist,currCoupling,query];
(* update the current distance with the optimal value function for the current coupling *)
currDist = UpdateDistanceValues[\[Lambda],ctmc,currDist,currCoupling,query];
(*DebugPrint[MatrixForm@currDist];*)
(* look for the next non-optimal transportation schedule *)
vtc = MatchingToChange[currCoupling,currDist,query];
];
active = ActivePairs[currDist,currCoupling,query];
(* Fix the distance for the current active pairs *)
DebugPrint["Current pair done... fixing values for the pairs ",active];

currDist=Fold[FixPair,currDist,active];
(* update the query list *)
sortedQuery = Select[sortedQuery,Not@FixedDistanceQ[currDist,#]&];
]];
DebugPrint["No more pairs left to process."];
DebugPrint["Total solved TPs: ",iterations];
DebugPrint["Execution Time: ",time, " sec"];

Switch[OptionValue[BDistCTMC,Output],
RulesForm, (#->getDistValue[currDist,#]&)/@queryLst,
VerboseForm, R[time, iterations, currDist],
_, currDist]
];


(* ::Subsection::Closed:: *)
(*Bisimilarity Quotient*)


BisimQuotientCTMC[S_,ctmc:CTMC[tau_,ell_,rho_,absQ_]]:=
Module[{Q,btau,bell,brho,babsQ,s},

Q=BisimClassesCTMC[S,ctmc];
Do[
s = First[c]; (* representant of the class c *)
bell[s]=ell[s]; (* labeling function *)
babsQ[s]=absQ[s]; (* absorbing state *)
If[Not[babsQ[s]],
(* transition function *)
btau[s]=Cases[(First[#]->Total[(#/.Append[tau[s],_->0]&)/@#])&/@Q,HoldPattern[_ -> v_/;v>0]];
(* exit-rate function *)
brho[s]= rho[s];
];
,{c,Q}];
{First/@Q,CTMC[btau,bell,brho,babsQ]}
];


BisimClassesCTMC[S_,ctmc:CTMC[_,_,_,_]]:=
\!\(TraditionalForm\`FixedPoint\)[(Union@@#&)/@ Gather[#,Intersection[#1,#2]!={}&]&,
Select[BDistCTMC[ctmc,1/2,Tuples[S,2],ConsistencyCheck->False,Verbose->False,Output->RulesForm],#[[2]]==0&][[All,1]]];


End[]


EndPackage[]
